#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdbool.h>
#include <math.h>
#include "simpletools.h"
#include "fdserial.h"
#include "config.h"
#include "gcode_struct.h"
#include "debug.h"

static char fileline[100];

static GCODE_STRUCT current_gcode;
static GCODE_STRUCT previous_gcode;

static bool inch_measurement;
static bool metric_measurement;
static bool absolute_positioning;
static bool relative_positioning;
static bool extruder_absolute;
static bool extruder_relative;

int calc_steps_required(double travel, int32_t steps_per_unit_of_measure)
{
	double steps;

	steps = fabs(travel) * (double)steps_per_unit_of_measure;

	return (uint32_t)floor(steps + 0.5);
}

void perform_gcode_struct_calcs()
{
	double travel;

	if(absolute_positioning == true && relative_positioning == false)
	{
		if(current_gcode.x_num > previous_gcode.x_num)
		{
			travel = current_gcode.x_num - previous_gcode.x_num;
			current_gcode.x_dir = X_INC_DIR;

			if(metric_measurement == true && inch_measurement == false)
			{
				current_gcode.x_total_steps = 
					calc_steps_required(travel, X_STEPS_PER_MM);
			}
			else if(metric_measurement == false && inch_measurement == true)
			{
				current_gcode.x_total_steps = 
					calc_steps_required(travel, X_STEPS_PER_IN);
			}
		}     
		else if(current_gcode.x_num < previous_gcode.x_num)
		{
			travel = previous_gcode.x_num - current_gcode.x_num;
			current_gcode.x_dir = X_DEC_DIR;

			if(metric_measurement == true && inch_measurement == false)
			{
				current_gcode.x_total_steps = 
					calc_steps_required(travel, X_STEPS_PER_MM);
			}
			else if(metric_measurement == false && inch_measurement == true)
			{
				current_gcode.x_total_steps = 
					calc_steps_required(travel, X_STEPS_PER_IN);
			}
		} 
		else
		{
			current_gcode.x_dir = previous_gcode.x_dir;
			current_gcode.x_total_steps = 0;
		}

		if(current_gcode.y_num > previous_gcode.y_num)
		{
			travel = current_gcode.y_num - previous_gcode.y_num;
			current_gcode.y_dir = Y_INC_DIR;

			if(metric_measurement == true && inch_measurement == false)
			{
				current_gcode.y_total_steps = 
					calc_steps_required(travel, Y_STEPS_PER_MM);
			}
			else if(metric_measurement == false && inch_measurement == true)
			{
				current_gcode.y_total_steps = 
					calc_steps_required(travel, Y_STEPS_PER_IN);
			}
		}     
		else if(current_gcode.y_num < previous_gcode.y_num)
		{
			travel = previous_gcode.y_num - current_gcode.y_num;
			current_gcode.y_dir = Y_DEC_DIR;

			if(metric_measurement == true && inch_measurement == false)
			{
				current_gcode.y_total_steps = 
					calc_steps_required(travel, Y_STEPS_PER_MM);
			}
			else if(metric_measurement == false && inch_measurement == true)
			{
				current_gcode.y_total_steps = 
					calc_steps_required(travel, Y_STEPS_PER_IN);
			}
		} 
		else
		{
			current_gcode.y_dir = previous_gcode.y_dir;
			current_gcode.y_total_steps = 0;
		}

		if(current_gcode.z_num > previous_gcode.z_num)
		{
			travel = current_gcode.z_num - previous_gcode.z_num;
			current_gcode.z_dir = Z_INC_DIR;

			if(metric_measurement == true && inch_measurement == false)
			{
				current_gcode.z_total_steps = 
					calc_steps_required(travel, Z_STEPS_PER_MM);
			}
			else if(metric_measurement == false && inch_measurement == true)
			{
				current_gcode.z_total_steps = 
					calc_steps_required(travel, Z_STEPS_PER_IN);
			}
		}     
		else if(current_gcode.z_num < previous_gcode.z_num)
		{
			travel = previous_gcode.z_num - current_gcode.z_num;
			current_gcode.z_dir = Z_DEC_DIR;

			if(metric_measurement == true && inch_measurement == false)
			{
				current_gcode.z_total_steps = 
					calc_steps_required(travel, Z_STEPS_PER_MM);
			}
			else if(metric_measurement == false && inch_measurement == true)
			{
				current_gcode.z_total_steps = 
					calc_steps_required(travel, Z_STEPS_PER_IN);
			}
		} 
		else
		{
			current_gcode.z_dir = previous_gcode.z_dir;
			current_gcode.z_total_steps = 0;
		}

		if(current_gcode.e_num > previous_gcode.e_num)
		{
			travel = current_gcode.e_num - previous_gcode.e_num;
			current_gcode.e_dir = E_INC_DIR;

			if(metric_measurement == true && inch_measurement == false)
			{
				current_gcode.e_total_steps = 
					calc_steps_required(travel, E_STEPS_PER_MM);
			}
			else if(metric_measurement == false && inch_measurement == true)
			{
				current_gcode.e_total_steps = 
					calc_steps_required(travel, E_STEPS_PER_IN);
			}
		}     
		else if(current_gcode.e_num < previous_gcode.e_num)
		{
			travel = previous_gcode.e_num - current_gcode.e_num;
			current_gcode.e_dir = E_DEC_DIR;

			if(metric_measurement == true && inch_measurement == false)
			{
				current_gcode.e_total_steps = 
					calc_steps_required(travel, E_STEPS_PER_MM);
			}
			else if(metric_measurement == false && inch_measurement == true)
			{
				current_gcode.e_total_steps = 
					calc_steps_required(travel, E_STEPS_PER_IN);
			}
		} 
		else
		{
			current_gcode.e_dir = previous_gcode.e_dir;
			current_gcode.e_total_steps = 0;
		}
	}	
	else if(absolute_positioning == false && relative_positioning == true)
	{
		if(current_gcode.x_num > 0)
		{
			travel = current_gcode.x_num;
			current_gcode.x_dir = X_INC_DIR;

			if(metric_measurement == true && inch_measurement == false)
			{
				current_gcode.x_total_steps = 
					calc_steps_required(travel, X_STEPS_PER_MM);
			}
			else if(metric_measurement == false && inch_measurement == true)
			{
				current_gcode.x_total_steps = 
					calc_steps_required(travel, X_STEPS_PER_IN);
			}
		}     
		else if(current_gcode.x_num < 0)
		{
			travel = current_gcode.x_num;
			current_gcode.x_dir = X_DEC_DIR;

			if(metric_measurement == true && inch_measurement == false)
			{
				current_gcode.x_total_steps = 
					calc_steps_required(travel, X_STEPS_PER_MM);
			}
			else if(metric_measurement == false && inch_measurement == true)
			{
				current_gcode.x_total_steps = 
					calc_steps_required(travel, X_STEPS_PER_IN);
			}
		} 
		else
		{
			current_gcode.x_dir = previous_gcode.x_dir;
			current_gcode.x_total_steps = 0;
		}

		if(current_gcode.y_num > 0)
		{
			travel = current_gcode.y_num;
			current_gcode.y_dir = Y_INC_DIR;

			if(metric_measurement == true && inch_measurement == false)
			{
				current_gcode.y_total_steps = 
					calc_steps_required(travel, Y_STEPS_PER_MM);
			}
			else if(metric_measurement == false && inch_measurement == true)
			{
				current_gcode.y_total_steps = 
					calc_steps_required(travel, Y_STEPS_PER_IN);
			}
		}     
		else if(current_gcode.y_num < 0)
		{
			travel = current_gcode.y_num;
			current_gcode.y_dir = Y_DEC_DIR;

			if(metric_measurement == true && inch_measurement == false)
			{
				current_gcode.y_total_steps = 
					calc_steps_required(travel, Y_STEPS_PER_MM);
			}
			else if(metric_measurement == false && inch_measurement == true)
			{
				current_gcode.y_total_steps = 
					calc_steps_required(travel, Y_STEPS_PER_IN);
			}
		} 
		else
		{
			current_gcode.y_dir = previous_gcode.y_dir;
			current_gcode.y_total_steps = 0;
		}

		if(current_gcode.z_num > 0)
		{
			travel = current_gcode.z_num;
			current_gcode.z_dir = Z_INC_DIR;

			if(metric_measurement == true && inch_measurement == false)
			{
				current_gcode.z_total_steps = 
					calc_steps_required(travel, Z_STEPS_PER_MM);
			}
			else if(metric_measurement == false && inch_measurement == true)
			{
				current_gcode.z_total_steps = 
					calc_steps_required(travel, Z_STEPS_PER_IN);
			}
		}     
		else if(current_gcode.z_num < 0)
		{
			travel = current_gcode.z_num;
			current_gcode.z_dir = Z_DEC_DIR;

			if(metric_measurement == true && inch_measurement == false)
			{
				current_gcode.z_total_steps = 
					calc_steps_required(travel, Z_STEPS_PER_MM);
			}
			else if(metric_measurement == false && inch_measurement == true)
			{
				current_gcode.z_total_steps = 
					calc_steps_required(travel, Z_STEPS_PER_IN);
			}
		} 
		else
		{
			current_gcode.z_dir = previous_gcode.z_dir;
			current_gcode.z_total_steps = 0;
		}

		if(current_gcode.e_num > 0)
		{
			travel = current_gcode.e_num;
			current_gcode.e_dir = E_INC_DIR;

			if(metric_measurement == true && inch_measurement == false)
			{
				current_gcode.e_total_steps = 
					calc_steps_required(travel, E_STEPS_PER_MM);
			}
			else if(metric_measurement == false && inch_measurement == true)
			{
				current_gcode.e_total_steps = 
					calc_steps_required(travel, E_STEPS_PER_IN);
			}
		}     
		else if(current_gcode.e_num < 0)
		{
			travel = current_gcode.e_num;
			current_gcode.e_dir = E_DEC_DIR;

			if(metric_measurement == true && inch_measurement == false)
			{
				current_gcode.e_total_steps = 
					calc_steps_required(travel, E_STEPS_PER_MM);
			}
			else if(metric_measurement == false && inch_measurement == true)
			{
				current_gcode.e_total_steps = 
					calc_steps_required(travel, E_STEPS_PER_IN);
			}
		} 
		else
		{
			current_gcode.e_dir = previous_gcode.e_dir;
			current_gcode.e_total_steps = 0;
		}
	}
}

void ini_gcode_struct()
{
	uint8_t gcode_char;
	uint8_t gcode_struct_element;
	uint8_t fileline_length;
	uint8_t value_counter;

	uint32_t counter;

	char g_char[3];
	char m_char[3];
	char t_char[1];
	char s_char[5];
	char p_char[5];
	char x_char[9];
	char y_char[8];
	char z_char[8];
	char e_char[8];
	char f_char[8];

	gcode_struct_element = 0;

	// Copy preexisting current_gcode.current_fileline to current_gcode.previous_fileline
	memset(current_gcode.previous_fileline, 0,
		strlen(current_gcode.previous_fileline));
	memcpy(&current_gcode.previous_fileline, &current_gcode.current_fileline,
		sizeof(current_gcode.current_fileline));

	// Clear any lingering data
	memset(current_gcode.current_fileline, 0, strlen(current_gcode.current_fileline));
	memset(current_gcode.valid, 0, strlen(current_gcode.valid));

	memcpy(&current_gcode.current_fileline, &fileline, sizeof(fileline));

	// Get the length of the fileline that is being parsed
	fileline_length = (uint8_t)strlen(fileline);

	// Fill current_gcode.valid with literal zeros
	for(counter = 0; counter < GCODE_ELEMENTS; counter++)
	{
		current_gcode.valid[counter] = '0';
	}

	value_counter = 0;

	//Parse file line into seperate field values.  Various fields will remain empty
	for(counter = 0; counter < fileline_length; counter++)
	{
		gcode_char = fileline[counter];

		// Exclude these characters from becoming part of are field entries
		if(gcode_char == 13 /*Carriage Return*/ || gcode_char == 10 /*Line Feed*/)
		{
			// Break free from this loop
			break;
		}

		else if(gcode_char == ' ')
		{
			value_counter = 0;
		}

		else if(gcode_char == 'G')
		{
			// Set current field element
			gcode_struct_element = 0;
			current_gcode.valid[0] = '1';
		}
		else if(gcode_char == 'M')
		{
			// Set current field element
			gcode_struct_element = 1;
			current_gcode.valid[1] = '1';
		}
		else if(gcode_char == 'T')
		{
			// Set current field element
			gcode_struct_element = 2;
			current_gcode.valid[2] = '1';
		}
		else if(gcode_char == 'S')
		{
			// Set current field element
			gcode_struct_element = 3;
			current_gcode.valid[3] = '1';
		}
		else if(gcode_char == 'P')
		{
			// Set current field element
			gcode_struct_element = 4;
			current_gcode.valid[4] = '1';
		}
		else if(gcode_char == 'X')
		{
			// Set current field element
			gcode_struct_element = 5;
			current_gcode.valid[5] = '1';
		}
		else if(gcode_char == 'Y')
		{
			// Set current field element
			gcode_struct_element = 6;
			current_gcode.valid[6] = '1';
		}
		else if(gcode_char == 'Z')
		{
			// Set current field element
			gcode_struct_element = 7;
			current_gcode.valid[7] = '1';
		}		
		else if(gcode_char == 'E')
		{
			// Set current field element
			gcode_struct_element = 8;
			current_gcode.valid[8] = '1';
		}
		else if(gcode_char == 'F')
		{
			// Set current field element
			gcode_struct_element = 9;
			current_gcode.valid[9] = '1';
		}
		else
		{
			if(gcode_struct_element == 0)
			{
				// G Field element character
				g_char[value_counter] = gcode_char;
				value_counter++;
			}
			else if(gcode_struct_element == 1)
			{
				// M Field element character
				m_char[value_counter] = gcode_char;
				value_counter++;
			}
			else if(gcode_struct_element == 2)
			{
				// T Field element character
				t_char[value_counter] = gcode_char;
				value_counter++;
			}
			else if(gcode_struct_element == 3)
			{
				// S Field element character
				s_char[value_counter] = gcode_char;
				value_counter++;
			}
			else if(gcode_struct_element == 4)
			{
				// P Field element character
				p_char[value_counter] = gcode_char;
				value_counter++;
			}
			else if(gcode_struct_element == 5)
			{
				// X Field element character
				x_char[value_counter] = gcode_char;
				value_counter++;
			}
			else if(gcode_struct_element == 6)
			{
				// Y Field element character
				y_char[value_counter] = gcode_char;
				value_counter++;
			}
			else if(gcode_struct_element == 7)
			{
				// Z Field element character
				z_char[value_counter] = gcode_char;
				value_counter++;
			}
			else if(gcode_struct_element == 8)
			{
				// E Field element character
				e_char[value_counter] = gcode_char;
				value_counter++;
			}
			else if(gcode_struct_element == 9)
			{
				// F Field element character
				f_char[value_counter] = gcode_char;
				value_counter++;
			}			
		}
	}

	current_gcode.g_num = atoi(g_char);
	current_gcode.m_num = atoi(m_char);
	current_gcode.t_num = atoi(t_char);
	current_gcode.s_num = atoi(s_char);
	current_gcode.p_num = atoi(p_char);

	current_gcode.x_num = atof(x_char);
	current_gcode.y_num = atof(y_char);
	current_gcode.z_num = atof(z_char);
	current_gcode.e_num = atof(e_char);
	current_gcode.f_num = atof(f_char);

	if(current_gcode.valid[0] == '1' && current_gcode.g_num == 20)
	{
		inch_measurement = true;
		metric_measurement = false;
	}

	if(current_gcode.valid[0] == '1' && current_gcode.g_num == 21)
	{
		inch_measurement = false;
		metric_measurement = true;
	}

	if(current_gcode.valid[0] == '1' && current_gcode.g_num == 90)
	{
		absolute_positioning = true;
		relative_positioning = false;
	}

	if(current_gcode.valid[0] == '1' && current_gcode.g_num == 91)
	{
		absolute_positioning = false;
		relative_positioning = true;
	}

	if(current_gcode.valid[1] == '1' && current_gcode.m_num == 82)
	{
		extruder_absolute = true;
		extruder_relative = false;
	}

	if(current_gcode.valid[1] == '1' && current_gcode.m_num == 83)
	{
		extruder_absolute = false;
		extruder_relative = true;
	}

	perform_gcode_struct_calcs();
}

int main(void)
{
	// Start full duplex serial for inter-propeller communication 
	fdserial *parser;
	parser = fdserial_open(INTER_PROP_RX, INTER_PROP_TX, INTER_PROP_MODE, INTER_PROP_BAUD);

	// Set debug level
	debug_flags |= DEBUG_ERROR;

	// Initialize fileline_number to zero, but increment this
	// variable for each valid fileline that is parsed
	current_gcode.fileline_number = 0;

	// Initialize positioning and measurement variables to false
	inch_measurement = false;
	metric_measurement = false;
	absolute_positioning = false;
	relative_positioning = false;
	extruder_absolute = false;
	extruder_relative = false;

	// Initialize previous_gcode GCODE_STRUCT members to zero
	previous_gcode.fileline_number = 0;

	previous_gcode.g_num = 0;
	previous_gcode.m_num = 0;
	previous_gcode.t_num = 0;
	previous_gcode.s_num = 0;
	previous_gcode.p_num = 0;
	previous_gcode.x_num = 0;
	previous_gcode.y_num = 0;
	previous_gcode.z_num = 0;
	previous_gcode.e_num = 0;
	previous_gcode.f_num = 0;

	previous_gcode.x_dir = 0; 
	previous_gcode.x_total_steps = 0;

	previous_gcode.y_dir = 0;
	previous_gcode.y_total_steps = 0;

	previous_gcode.z_dir = 0;
	previous_gcode.z_total_steps = 0;

	previous_gcode.e_dir = 0;
	previous_gcode.e_total_steps = 0;

	// Verify that SD card pins are defined
	#if defined PMC_IO1_DO && defined PMC_CLK && defined PMC_IO0_DI && defined PMC_SD_CS

	// Mount SD card
	sd_mount(PMC_IO1_DO, PMC_CLK, PMC_IO0_DI, PMC_SD_CS);

	FILE * pFile;

	// Open up a file on the SD card
	pFile = fopen("DEBUG.TXT","r");

	if(pFile == NULL)
	{
		if(DEBUG_ERROR && (debug_flags & DEBUG_ERROR))
		{
			print("ERROR - The file could not be opened!");
		}
	}
	else
	{
		// Recursively grab filelines
		while(fgets(fileline, 100, pFile))
		{
			// Count the filelines as we go along
			current_gcode.fileline_number++;

			// Exclude lines that begin with certain characters
			// E.G. Remove newlines, carriage returns, and lines
			// that begin with a semi-colon.
			if(fileline[0] != 59 && fileline[0] != 13 && fileline[0] != 10)
			{
				// Fill the GCODE_STRUCT with the data
				// obtained from the fileline
				ini_gcode_struct();

				// Copy GCODE_STRUCT into char array for sending
				char send_struct_array[sizeof(GCODE_STRUCT)];
				memcpy(&send_struct_array, &current_gcode, sizeof(GCODE_STRUCT));				

				// proceed the struct with a '<' for STX
				// STX = start of text
				// STX can be anything that will not be included in your data 
				// STX must be the same on the receiving end
				fdserial_txChar(parser, '<');

				for(int i = 0; i < sizeof(GCODE_STRUCT); i++)
				{
					while(fdserial_txEmpty(parser) == 0){}

					fdserial_txChar(parser, send_struct_array[i]);
				}

				// follow the struct with a '>' for ETX
				// ETX = end of text
				// ETX can be anything that will not be included in your data 
				// ETX must be the same on the receiving end
				fdserial_txChar(parser, '>');

				// Still need to determine how to handle the following line
				//while(gcode_struct_received == false){}

				// Still need to determine how to handle the following line
				//gcode_struct_received = false;
			}

			// Copy the current GCODE_STRUCT into the previous one
			memcpy(&previous_gcode, &current_gcode, sizeof(GCODE_STRUCT));

			// Clear any lingering data from the fileline variable
			memset(fileline, 0, strlen(fileline));
		}

		// Close up the file on the SD card
		fclose(pFile);
	}
 
	#else
    
	if(DEBUG_ERROR && (debug_flags & DEBUG_ERROR))
	{
		print("ERROR - The proper inputs and outputs must be defined for SD card usage!");
	}
    
	#endif

	return 0;
}